"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var StringUtils_1 = require("./StringUtils");
var NodeUtils = /** @class */ (function () {
    function NodeUtils() {
    }
    NodeUtils.getValidTextNode = function (node) {
        if (node.nodeType === 3 && /\S/.test(node.data)) {
            return node;
        }
        else if (node.nodeType === 1 && !NodeUtils.isSkippable(node)) {
            for (var i = 0, len = node.childNodes.length; i < len; ++i) {
                var valid = NodeUtils.getValidTextNode(node.childNodes[i]);
                if (valid) {
                    return valid;
                }
            }
        }
        return NodeUtils.getValidTextNode(node.nextSibling);
    };
    NodeUtils.getValidTextNodes = function (node) {
        if (node.nodeType === 3 && /\S/.test(node.data)) {
            return [node];
        }
        else if (node.nodeType === 1 && !NodeUtils.isSkippable(node)) {
            return Array.from(node.childNodes).reduce(function (nodes, child) {
                return nodes.concat(NodeUtils.getValidTextNodes(child));
            }, []);
        }
        return [];
    };
    NodeUtils.isSkippable = function (node) {
        return NodeUtils.elementsToSkip.test(node, null);
    };
    NodeUtils.split = function (text, offset) {
        var rp = text.splitText(offset);
        rp.startPosition = text.startPosition + StringUtils_1.StringUtils.compact(text.data).length;
        rp.endPosition = text.endPosition;
        text.endPosition = rp.startPosition - 1;
        return rp;
    };
    NodeUtils.elementsToSkip = {
        elements: [
            'applet',
            'area',
            'base',
            'basefont',
            'bdo',
            'button',
            'frame',
            'frameset',
            'iframe',
            'head',
            'hr',
            'img',
            'input',
            'link',
            'map',
            'meta',
            'noframes',
            'noscript',
            'optgroup',
            'option',
            'param',
            'script',
            'select',
            'style',
            'textarea',
            'title',
        ],
        /**
         * @param {Element} node
         * @param {string[]|null} list
         * @return {boolean}
         */
        test: function (node, list) {
            var elements = list || this.elements;
            return elements.indexOf(node.tagName.toLowerCase()) > -1;
        },
    };
    return NodeUtils;
}());
exports.NodeUtils = NodeUtils;
//# sourceMappingURL=NodeUtils.js.map