(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.texton = {})));
}(this, (function (exports) { 'use strict';

  var StringUtils = /** @class */ (function () {
      function StringUtils() {
      }
      /**
       * the position of the nth occurrence of searchString
       * @param {string} source
       * @param {string} searchString
       * @param {number} nth = 1
       * @return {number} start position of the nth occurrence of <em>searchString</em>
       */
      StringUtils.indexOf = function (source, searchString, nth) {
          if (nth === void 0) { nth = 1; }
          var times = 0;
          var index = -2;
          while (times < nth && index !== -1) {
              index = source.indexOf(searchString, index + 1);
              times++;
          }
          return index;
      };
      StringUtils.indexOfNthNotEmptyChar = function (source, nth) {
          var i = 0;
          for (var n = -1; n <= nth && i < source.length; i++) {
              if (!/\s+/m.test(source[i])) {
                  n++;
                  if (n === nth) {
                      return i;
                  }
              }
          }
          return -1;
      };
      /**
       * the occurrences(include the positions) of searchString
       * @param {string} source
       * @param {string} searchString
       * @returns {Array<IOccurrence>}
       */
      StringUtils.find = function (source, searchString) {
          var times = 0;
          var position = -2;
          var result = [];
          if (searchString.length < 1) {
              return result;
          }
          while (position !== -1) {
              position = source.indexOf(searchString, position + 1);
              if (position < 0) {
                  break;
              }
              result.push({ nth: ++times, position: position });
          }
          return result;
      };
      /**
       * @return {string} compacted string
       */
      StringUtils.compact = function (source) {
          return source.replace(/\s+/gm, '');
      };
      /**
       * @return {string} trimmed string
       */
      StringUtils.trim = function (source) {
          return source.replace(/(^\s*)|(\s*$)/g, '');
      };
      return StringUtils;
  }());

  var NodeUtils = /** @class */ (function () {
      function NodeUtils() {
      }
      NodeUtils.getValidTextNode = function (node) {
          if (node.nodeType === 3 && /\S/.test(node.data)) {
              return node;
          }
          else if (node.nodeType === 1 && !NodeUtils.isSkippable(node)) {
              for (var i = 0, len = node.childNodes.length; i < len; ++i) {
                  var valid = NodeUtils.getValidTextNode(node.childNodes[i]);
                  if (valid) {
                      return valid;
                  }
              }
          }
          return NodeUtils.getValidTextNode(node.nextSibling);
      };
      NodeUtils.getValidTextNodes = function (node) {
          if (node.nodeType === 3 && /\S/.test(node.data)) {
              return [node];
          }
          else if (node.nodeType === 1 && !NodeUtils.isSkippable(node)) {
              return Array.from(node.childNodes).reduce(function (nodes, child) {
                  return nodes.concat(NodeUtils.getValidTextNodes(child));
              }, []);
          }
          return [];
      };
      NodeUtils.isSkippable = function (node) {
          return NodeUtils.elementsToSkip.test(node, null);
      };
      NodeUtils.split = function (text, offset) {
          var rp = text.splitText(offset);
          rp.startPosition = text.startPosition + StringUtils.compact(text.data).length;
          rp.endPosition = text.endPosition;
          text.endPosition = rp.startPosition - 1;
          return rp;
      };
      NodeUtils.elementsToSkip = {
          elements: [
              'applet',
              'area',
              'base',
              'basefont',
              'bdo',
              'button',
              'frame',
              'frameset',
              'iframe',
              'head',
              'hr',
              'img',
              'input',
              'link',
              'map',
              'meta',
              'noframes',
              'noscript',
              'optgroup',
              'option',
              'param',
              'script',
              'select',
              'style',
              'textarea',
              'title',
          ],
          /**
           * @param {Element} node
           * @param {string[]|null} list
           * @return {boolean}
           */
          test: function (node, list) {
              var elements = list || this.elements;
              return elements.indexOf(node.tagName.toLowerCase()) > -1;
          },
      };
      return NodeUtils;
  }());

  var XSelection$$1 = /** @class */ (function () {
      function XSelection$$1(range, xDocument) {
          this.range = range;
          this.xDocument = xDocument;
      }
      /**
       * create {@link XSelection} from native {@link Selection}
       * @param selection the native `Selection` from which to create the `XSelection` instance.
       * @param xDoc the context of the to be created `XSelection` instance.
       */
      XSelection$$1.fromSelection = function (selection, xDoc) {
          var range = XSelection$$1.rangeFromSelection(selection, xDoc);
          return XSelection$$1.fromRange(range, false, xDoc);
      };
      /**
       * create {@link XSelection} from {@link ITextRange} pojo;
       * {@link ITextRange} specifies the start point and end point
       * <pre>
       * export interface ITextRange {
       *   from: { text: string, nth: number }, // means starting from the `nth` `text`
       *   to: { text: string, nth: number }, // means ending at the `nth` `text`
       * }
       * </pre>
       * @param textRange the serialized pojo from which to create the `XSelection` instance.
       * @param optSelect true to create explicit selection(be selected visibly from browser), false otherwise
       * @param xDoc the context of the to be created `XSelection` instance.
       */
      XSelection$$1.fromTextRange = function (textRange, optSelect, xDoc) {
          if (optSelect === void 0) { optSelect = false; }
          var range = XSelection$$1.rangeFromTextRange(textRange, xDoc);
          return XSelection$$1.fromRange(range, optSelect, xDoc);
      };
      /**
       * create {@link XSelection} from {@link ITextRange} pojo
       * @param text the text content to be selected.
       * @param nth together with {@param text} specifies the exact {@code nth} {@code text} will be used to create the selection.
       * @param optSelect true to create explicit selection(be selected visibly from browser), false otherwise
       * @param xDoc the context of the to be created `XSelection` instance.
       */
      XSelection$$1.fromText = function (text, nth, optSelect, xDoc) {
          if (nth === void 0) { nth = 1; }
          if (optSelect === void 0) { optSelect = false; }
          var range = XSelection$$1.rangeFromText(text, nth, xDoc);
          return XSelection$$1.fromRange(range, optSelect, xDoc);
      };
      /**
       * create {@link XSelection} from native {@link Range}
       * @param range the text content to be selected.
       * @param optSelect true to create selection explicitly(be selected visibly from browser), false otherwise
       * @param xDoc the context of the to be created `XSelection` instance.
       */
      XSelection$$1.fromRange = function (range, optSelect, xDoc) {
          if (optSelect === void 0) { optSelect = false; }
          if (range) {
              if (optSelect) {
                  var selection = xDoc.win.getSelection();
                  if (selection) {
                      selection.removeAllRanges();
                      selection.addRange(range);
                  }
              }
              return new XSelection$$1(range, xDoc);
          }
          return null;
      };
      XSelection$$1.rangeFromSelection = function (selection, xDoc) {
          if (selection.rangeCount > 0 && !selection.isCollapsed) {
              return selection.getRangeAt(0);
          }
          return null;
      };
      XSelection$$1.rangeFromTextRange = function (textRange, xDoc) {
          var from = XSelection$$1.rangeFromText(textRange.from.text, textRange.from.nth, xDoc);
          var to = XSelection$$1.rangeFromText(textRange.to.text, textRange.to.nth, xDoc);
          if (from && to) {
              var range = xDoc.doc.createRange();
              range.setStart(from.startContainer, from.startOffset);
              range.setEnd(to.endContainer, to.endOffset);
              return range;
          }
          return null;
      };
      /**
       *  return an empty range object that has both of its boundary points positioned at
       *  the beginning of the document when no parameters are given, else return an range
       *  object that contains the nth occurrence of the specified text
       *
       * @param {string} text the base text
       * @param {number} nth = 1 the nth index
       * @param {XDocument} xDoc
       * @return {Range} the range created that contains the specified text
       * @private
       */
      XSelection$$1.rangeFromText = function (text, nth, xDoc) {
          if (nth === void 0) { nth = 1; }
          var nodes = xDoc.nodes;
          var cText = StringUtils.compact(text);
          var index = StringUtils.indexOf(xDoc.text, cText, nth || 1);
          if (index > -1) {
              var sPos_1 = index + nodes[0].startPosition; // include the first character
              var ePos_1 = sPos_1 + cText.length; // exclude the last character
              var sIndex = nodes.findIndex(function (node) { return node.endPosition >= sPos_1; });
              var eIndex = nodes.findIndex(function (node) { return node.startPosition >= ePos_1; }) - 1;
              var sContainer = nodes[sIndex];
              var eContainer = eIndex >= 0 ? nodes[eIndex] : nodes[nodes.length - 1];
              if (sContainer) {
                  var sOffset = StringUtils.indexOfNthNotEmptyChar(sContainer.data, sPos_1 - sContainer.startPosition);
                  var eOffset = StringUtils.indexOfNthNotEmptyChar(eContainer.data, ePos_1 - eContainer.startPosition - 1);
                  var range = xDoc.doc.createRange();
                  range.setStart(sContainer, sOffset);
                  range.setEnd(eContainer, eOffset + 1);
                  return range;
              }
          }
          return null;
      };
      Object.defineProperty(XSelection$$1.prototype, "texts", {
          /**
           * the text nodes contained in the selection.
           * <em>
           * text node will be splatted into 2 parts if the selection is start from/end in a internal
           * position of a text node
           * </em>
           */
          get: function () {
              if (this.xTexts) {
                  return this.xTexts;
              }
              var sContainer = this.range.startContainer;
              var eContainer = this.range.endContainer;
              var sOffset = this.range.startOffset;
              var eOffset = this.range.endOffset;
              if (sContainer.nodeType !== 3) {
                  sContainer = NodeUtils.getValidTextNode(sContainer);
              }
              if (eContainer.nodeType !== 3) {
                  eContainer = NodeUtils.getValidTextNode(eContainer);
              }
              var bTexts = this.xDocument.nodes;
              var si = bTexts.indexOf(sContainer);
              var srp = NodeUtils.split(sContainer, sOffset);
              if (srp.length > 0) {
                  bTexts.splice(si + 1, 0, srp);
              }
              if (eContainer === sContainer) {
                  eContainer = srp;
                  eOffset = eOffset - sOffset;
              }
              var ei = bTexts.indexOf(eContainer);
              var erp = NodeUtils.split(eContainer, eOffset);
              if (erp.length > 0) {
                  bTexts.splice(ei + 1, 0, erp);
              }
              this.xTexts = bTexts.slice(si + 1, ei + 1);
              this.range.setStart(bTexts[si + 1], 0);
              this.range.setEnd(bTexts[ei], bTexts[ei].data.length);
              return this.xTexts;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * get the exact "occurrence" of this selection.
       * <em>
       *   the selected text content may appear >=1 times in this page, and the selection is it's
       *   {@code occurrence.nth} time. and the start char of this selection is the
       *   {@code occurrence.position}th char of the whole text content in this page.
       * </em>
       *
       */
      XSelection$$1.prototype.getOccurrence = function () {
          var cText = StringUtils.compact(this.range.toString());
          if (cText.length < 1) {
              return exports.IOccurrence.None;
          }
          var eContainer = this.range.endContainer;
          if (eContainer.nodeType !== 3) {
              eContainer = NodeUtils.getValidTextNode(eContainer);
          }
          var offset = this.range.endOffset;
          var cHead = StringUtils.compact(eContainer.substringData(0, offset));
          var cLength = cHead ? cHead.length : 0;
          var content = this.xDocument.text.substr(0, eContainer.startPosition + cLength);
          var occurrences = StringUtils.find(content, cText);
          if (occurrences && occurrences.length > 0) {
              return occurrences[occurrences.length - 1];
          }
          return exports.IOccurrence.None;
      };
      /**
       * get the text range of the selection in this page.
       * <pre>
       * export interface ITextRange {
       *   from: { text: string, nth: number }, // means starting from the `nth` `text`
       *   to: { text: string, nth: number }, // means ending at the `nth` `text`
       * }
       * </pre>
       * that is, the text content of this selection starts with {@code textRange.from.text} and
       * ends with {@code textRange.to.text}. and {@code textRange.from.text} is its {@code textRange.from.nth}
       * occurrence, {@code textRange.to.text} is its {@code textRange.to.nth} occurrence.
       * e.g.
       * for text
       * <span>hello world, hello miller, hello world, hello miller, hello miller</span>
       * <pre>
       * {
       *   from:{text: 'hello', nth: 2}
       *   to:{text: 'miller', nth: 3}
       * }
       * means, the selection starts from the 2nd `hello` and ends with 3rd `miller`. that is:
       * <span>hello world, <selected>hello miller, hello world, hello miller, hello miller<selected>, hello world<span>
       * </pre>
       *
       * @param len
       */
      XSelection$$1.prototype.getTextRange = function (len) {
          if (len === void 0) { len = 5; }
          var cText = StringUtils.compact(this.range.toString());
          if (cText.length < 1) {
              return exports.ITextRange.None;
          }
          var eText = cText.length > 5 ? cText.substr(cText.length - 5, 5) : cText;
          var sText = cText.length > 5 ? cText.substr(0, 5) : cText;
          var eContainer = this.range.endContainer;
          if (eContainer.nodeType !== 3) {
              eContainer = NodeUtils.getValidTextNode(eContainer);
          }
          var eOffset = this.range.endOffset;
          var eHead = StringUtils.compact(eContainer.substringData(0, eOffset));
          var eLength = eHead ? eHead.length : 0;
          var eContent = this.xDocument.text.substr(0, eContainer.startPosition + eLength);
          var eOccurs = StringUtils.find(eContent, eText);
          var sContainer = this.range.startContainer;
          if (sContainer.nodeType !== 3) {
              sContainer = NodeUtils.getValidTextNode(sContainer);
          }
          var sOffset = this.range.startOffset;
          var sHead = StringUtils.compact(sContainer.substringData(0, sOffset));
          var sLength = sHead ? sHead.length : 0;
          var sContent = this.xDocument.text.substr(0, sContainer.startPosition + sLength + sText.length);
          var sOccurs = StringUtils.find(sContent, sText);
          if (sOccurs && sOccurs.length > 0 && eOccurs && eOccurs.length > 0) {
              return {
                  from: { text: sText, nth: sOccurs[sOccurs.length - 1].nth },
                  to: { text: eText, nth: eOccurs[eOccurs.length - 1].nth }
              };
          }
          return exports.ITextRange.None;
      };
      /**
       * get the selected text content
       * @param compact true to remove the white spaces. false otherwise
       */
      XSelection$$1.prototype.getContent = function (compact) {
          if (compact === void 0) { compact = false; }
          if (compact) {
              return StringUtils.compact(this.range.toString());
          }
          return this.range.toString();
      };
      Object.defineProperty(XSelection$$1.prototype, "selection", {
          /**
           * get the native selection
           */
          get: function () {
              var selection = this.xDocument.win.getSelection();
              if (selection.rangeCount < 1) {
                  selection.addRange(this.range);
              }
              return selection;
          },
          enumerable: true,
          configurable: true
      });
      /**
       * clear the selection
       */
      XSelection$$1.prototype.empty = function () {
          this.range.collapse();
          var selection = this.xDocument.win.getSelection();
          selection.collapseToStart();
          selection.empty && selection.empty();
          selection.removeAllRanges && selection.removeAllRanges();
      };
      /**
       * check if the selection is cleared or not.
       */
      XSelection$$1.prototype.isEmpty = function () {
          var selection = this.selection;
          return selection.rangeCount < 1 || selection.isCollapsed || this.range.collapsed;
      };
      XSelection$$1.prototype.cancel = function () {
          var texts = this.xTexts;
          if (!texts || texts.length < 1) {
              return;
          }
          var first = texts[0];
          var last = texts[texts.length - 1];
          if (first && first.parentNode) {
              first.parentNode.normalize();
          }
          if (last && last.parentNode) {
              last.parentNode.normalize();
          }
          var bTexts = this.xDocument.nodes;
          var fi = bTexts.indexOf(first);
          var li = bTexts.indexOf(last);
          if (li < bTexts.length - 1 && last.nextSibling === bTexts[li + 1]) {
              last.endPosition = bTexts[li + 1].endPosition;
              bTexts.splice(li + 1, 1);
          }
          if (fi > 0 && first.previousSibling === bTexts[fi - 1]) {
              bTexts[fi - 1].endPosition = first.endPosition;
              bTexts.splice(fi, 1);
          }
      };
      return XSelection$$1;
  }());

  var XDocument = /** @class */ (function () {
      function XDocument(root) {
          this.selection = null;
          this.root = root;
          this.doc = (root instanceof Document ? root : root.ownerDocument);
          this.win = this.doc.defaultView;
          var _a = this.prepare(root), text = _a.text, nodes = _a.nodes;
          this.text = text;
          this.nodes = nodes;
      }
      XDocument.from = function (root) {
          if (root === void 0) { root = document.body; }
          return new XDocument(root);
      };
      /**
       * @see {@code XSelection.fromText}
       */
      XDocument.prototype.select = function (optText, optNth, optSelect) {
          if (optNth === void 0) { optNth = 1; }
          if (optSelect === void 0) { optSelect = false; }
          if (optText && optText.trim()) {
              if (this.selection && !this.selection.isEmpty()) {
                  this.selection.cancel();
                  this.selection.selection.empty();
              }
              this.selection = this.fromText(optText.trim(), optNth);
          }
          else {
              var selection = this.win.getSelection();
              if (selection) {
                  this.selection = this.fromSelection(selection);
              }
          }
          return this.selection;
      };
      /**
       * @see {@code XSelection.fromSelection}
       */
      XDocument.prototype.fromSelection = function (selection) {
          return this.selection = XSelection$$1.fromSelection(selection, this);
      };
      /**
       * @see {@code XSelection.fromText}
       */
      XDocument.prototype.fromText = function (text, nth, optSelect) {
          if (nth === void 0) { nth = 1; }
          if (optSelect === void 0) { optSelect = false; }
          return this.selection = XSelection$$1.fromText(text, nth, optSelect, this);
      };
      /**
       * @see {@code XSelection.fromTextRange}
       */
      XDocument.prototype.fromTextRange = function (range, optSelect) {
          if (optSelect === void 0) { optSelect = false; }
          return this.selection = XSelection$$1.fromTextRange(range, optSelect, this);
      };
      XDocument.prototype.prepare = function (root) {
          var texts = NodeUtils.getValidTextNodes(root);
          var position = 0;
          var xText = '';
          texts.forEach(function (node) {
              node.startPosition = position;
              var cText = StringUtils.compact(node.data);
              xText += cText;
              position += cText.length;
              node.endPosition = position - 1;
          });
          return { text: xText, nodes: texts };
      };
      XDocument.prototype.normalize = function () {
          var _this = this;
          this.root.normalize();
          var end = 0;
          this.nodes.reverse().forEach(function (node, index) {
              if (!node.parentNode) {
                  end = node.endPosition;
                  _this.nodes.splice(_this.nodes.length - index, 1);
              }
              else if (end > 0) {
                  node.endPosition = end;
              }
          });
      };
      return XDocument;
  }());

  (function (IOccurrence) {
      IOccurrence.None = {
          nth: -1,
          position: -1,
      };
  })(exports.IOccurrence || (exports.IOccurrence = {}));
  (function (ITextRange) {
      ITextRange.None = {
          from: { text: '', nth: -1 },
          to: { text: '', nth: -1 },
      };
      function toSimpleString(range) {
          return JSON.stringify([range.from.nth + ":" + range.from.text, range.to.nth + ":" + range.to.text]);
      }
      ITextRange.toSimpleString = toSimpleString;
      function fromSimpleString(str) {
          var simple = JSON.parse(str);
          var from = simple[0].split(':', 2);
          var to = simple[1].split(':');
          return {
              from: { text: from[1], nth: parseInt(from[0], 10) },
              to: { text: to[1], nth: parseInt(to[0], 10) }
          };
      }
      ITextRange.fromSimpleString = fromSimpleString;
  })(exports.ITextRange || (exports.ITextRange = {}));

  exports.XSelection = XSelection$$1;
  exports.XDocument = XDocument;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=texton.umd.js.map
